# utils.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import os
import logging

from typing import Callable, Optional, TypedDict

from contextlib import suppress

from gi.repository import GLib, Xdp, GLib, Gio, Gtk, GObject


class RefineSettingsSource(TypedDict):
    """Dictionary for storing important `Gio.Settings`-related data."""

    settings: Optional[Gio.Settings]
    schema: Optional[Gio.SettingsSchema]


class RefineSettings:
    """Metaclass providing decorators for common GObject properties."""

    @staticmethod
    def value(func: Callable) -> Callable:
        """`@Refine.value` decorator implementing the `GObject:value` property."""

        @GObject.Property(type=str, default="")
        def value(self) -> str:
            return self._value

        @value.setter  # type: ignore [no-redef]
        def value(self, value) -> None:
            if not self.is_valid_setting:
                return

            func(self, value)

            self._value = value

        return value

    @staticmethod
    def key(func: Callable) -> Callable:
        """`@Refine.key` decorator implementing the `GObject:key` property."""

        @GObject.Property(
            type=str,
            default="",
            flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.EXPLICIT_NOTIFY,
        )
        def key(self) -> str:
            return self._key

        @key.setter  # type: ignore [no-redef]
        def key(self, key) -> None:
            if not self.is_valid_setting:
                return

            try:
                if not self.schema.has_key(key):
                    raise TypeError

                func(self, key)
            except Exception:
                self.is_valid_setting = False
                logging.warning(f"Key “{key}” does not exist. Skipping…")

            self._key = key
            self.notify("key")

        return key

    @staticmethod
    def schema_id(func: Callable) -> Callable:
        """`@Refine.schema_id` decorator implementing the `GObject:schema-id` property."""

        @GObject.Property(type=str, default="")
        def schema_id(self) -> str:
            return self._schema_id

        @schema_id.setter  # type: ignore [no-redef]
        def schema_id(self, schema_id: str) -> None:
            toplevels = Gtk.Window.get_toplevels()

            if not (window := toplevels.get_item(0)):
                return

            application = window.get_application()

            try:
                self.settings = application.settings[schema_id]["settings"]
                self.schema = application.settings[schema_id]["schema"]
                self.is_valid_setting = bool(self.settings)
                return
            except KeyError:
                pass

            try:
                self.schema = application.source.lookup(schema_id, True)

                if not self.schema:
                    raise TypeError
            except (AttributeError, TypeError):
                self.is_valid_setting = False
                application.settings[schema_id] = {"settings": None, "schema": None}
                logging.warning(f"Schema ID “{schema_id}” does not exist. Skipping…")
                return

            self.settings = Gio.Settings.new(schema_id)
            application.settings[schema_id] = {
                "settings": self.settings,
                "schema": self.schema,
            }

            self._schema_id = schema_id

        return schema_id


def get_user_config_dir() -> str:
    """Retrieve the user's config directory."""

    if not Xdp.Portal.running_under_sandbox():
        return GLib.get_user_config_dir()

    return os.environ.get(
        "HOST_XDG_CONFIG_HOME", os.path.join(GLib.get_home_dir(), "." + "config")
    )


def get_user_data_dir() -> str:
    """Retrieve the user's data directory."""

    if not Xdp.Portal.running_under_sandbox():
        return GLib.get_user_data_dir()

    return os.environ.get(
        "HOST_XDG_DATA_HOME",
        os.path.join(GLib.get_home_dir(), "." + "local", "share"),
    )


def get_system_data_dirs() -> list[str]:
    """Retrieve the system's data directory."""

    data_dirs = GLib.get_system_data_dirs()

    with suppress(ValueError):
        data_dirs.remove(os.path.join(os.sep, "app", "share"))

    return data_dirs


def get_debug_info() -> str:
    """Retrieve important information for debugging."""

    environment_variables = [
        "XDG_DATA_DIRS",
        "XDG_CONFIG_HOME",
        "XDG_DATA_HOME",
        "HOST_XDG_CONFIG_HOME",
        "HOST_XDG_DATA_HOME",
    ]

    output = ""
    for environment_variable in environment_variables:
        output += f"{environment_variable}: “{GLib.getenv(environment_variable)}”\n"

    return str(output)


def create_schema_source() -> Optional[Gio.SettingsSchemaSource]:
    """Create a schema source based on the user's system data directory."""

    for data_dir in GLib.get_system_data_dirs():
        try:
            source = Gio.SettingsSchemaSource.new_from_directory(
                os.path.join(data_dir, "glib-2.0", "schemas"), None, True
            )
            return source
        except GLib.GError:  # type: ignore [misc]
            pass

    return None


def __state_flags_changed(
    widget, _flags: Gtk.StateFlags, revealer: Gtk.Revealer
) -> None:
    flags = widget.get_state_flags()

    is_prelight = flags & (Gtk.StateFlags.PRELIGHT)
    is_activated = flags & (Gtk.StateFlags.SELECTED | Gtk.StateFlags.CHECKED)
    is_focused = flags & Gtk.StateFlags.FOCUSED & Gtk.StateFlags.FOCUS_VISIBLE

    revealer.set_reveal_child(not (is_prelight or is_activated or is_focused))


def set_up_separator_revealer(
    revealer: Gtk.Revealer, widget: Gtk.Widget, widget_property: str
) -> None:
    """Initialize and set up the `Gtk.Revealer`."""

    revealer.set_reveal_child(True)
    widget.connect("state-flags-changed", __state_flags_changed, revealer)
    widget.bind_property(
        "visible", revealer, "visible", GObject.BindingFlags.SYNC_CREATE
    )
    widget.bind_property(
        "opacity", revealer, "opacity", GObject.BindingFlags.SYNC_CREATE
    )
